#include <iostream>
#include <list>
#include <queue>
#include <vector>
#include <stack>
using namespace std;

#include "Position.hpp"
#include "Echiquier.hpp"
#include "UtilsEchiquier.cpp"
using namespace UtilsEchiquier;
#include "UtilsList.cpp"
using namespace UtilsList;

// Tableau de deplacements: mouvements standard
const Position arrDelta0[] = {
  Position(-2,-1), Position(-1,-2),
  Position( 1,-2), Position( 2,-1),
  Position( 2, 1), Position( 1, 2),
  Position(-1, 2), Position(-2, 1)
};

// Tableau de deplacements: autres mouvements
const Position arrDelta1[] = {
  Position( 0,-3), Position( 2,-2),
  Position( 3, 0), Position( 2, 2),
  Position( 0, 3), Position(-2, 2),
  Position(-3, 0), Position(-2,-2)
};

// Vecteurs bases sur les tableaux
typedef vector<Position> CVDeplacements;
const CVDeplacements delta0(&arrDelta0[0], &arrDelta0[8]);
const CVDeplacements delta1(&arrDelta1[0], &arrDelta1[8]);

// Surdefinit l'operation + sur les Position
Position operator+(const Position& pos1, const Position& pos2)
{
  return Position(pos1.getX()+pos2.getX(), pos1.getY()+pos2.getY());
}

// Operation + sur les Position
Position plusPos(const Position& pos1, const Position& pos2)
{
  return Position(pos1.getX()+pos2.getX(), pos1.getY()+pos2.getY());
}

/**
   Parcourt en largeur d'un Echiquier a partir d'une Position
   @param[in,out] cb - un Echiquier
   @param[in] delta - delta de deplacements
   @param[in] depart - Position de depart
*/
void bfsCV(Echiquier& cb, const CVDeplacements& delta, const Position& depart)
{
  // Reinitialise l'echiquier
  cb.reset(0);

  // Se positionne au point de depart et fixe sa valeur (marque)
  cb.set(depart, 1);

  // Definit la file des positions a examiner et enfile la
  // position de depart
  queue<Position> q;
  q.push(depart);

  // Parcourt tantque possible
  while( not q.empty() )
  {
    // Recupere la "prochaine" position (x,y) de la file
    // pour la traiter puis la defile
    Position pos = q.front();
    q.pop();

    // Recupere sa valeur sur l'echiquier
    int val = cb.get(pos);

    // Marque toutes les cases accessibles a partir d'elles
    for (unsigned nv = 0; nv < delta.size(); ++nv)
    {
      Position nextpos = pos + delta[nv];
      // Si nextpos est dans le domaine et non encore marquee,
      // alors visite la cellule et l'enfile dans la queue
      if (cb.contient(nextpos) && cb.get(nextpos) == 0)
      {
        cb.set(nextpos, val+1);
        q.push(nextpos);
      }
    }
  }
}

/**
  Affiche la liste des cellules accessibles et calcule
  la profondeur de la visite
  @param[in] cb - un Echiquier
*/
void afficherListeBFS(const Echiquier& cb)
{
  list<Position> cells;
  int ncoups = 0;
  for (unsigned iy = 0; iy < cb.sizeN(); ++iy)
  {
    for (unsigned ix = 0; ix < cb.sizeN(); ++ix)
    {
      Position c(ix, iy);
      if ( cb.get(ix, iy) != 0 )
      {
        cells.push_back(c);
        ncoups = max(ncoups, cb.get(c));
      }
    }
  }
  cout<<cells.size()<<" positions accessibles en "
      <<ncoups<<" coups:"<<endl;
  afficherList(cells);
}

/**
  @test
*/
void test_BFS()
{
  // Saisit des donnees
  cout<<"Dimension n de l'echiquier? ";
  unsigned n;
  cin >> n;
  Echiquier cb(n);
  cout<<"Mouvements standard (0 == oui)? ";
  unsigned rep;
  cin >> rep;
  cout<<"Position (x,y) de depart? ";
  Position cell;
  cin >> cell;
  // Lance la traversee en largeur
  if (rep == 0)
  {
    bfsCV(cb, delta0, cell);
  }
  else
  {
    bfsCV(cb, delta1, cell);
  }
  // Affiche le resultat
  cout<<"Traversee en largeur: "<<endl;
  afficherEchiquier(cb, 3);
  afficherListeBFS(cb);
}

/**
  Traversee en profondeur d'un Echiquier
  @param[in,out] cb - un Echiquier
  @param[in] delta - delta de deplacements
  @param[in] pos - la Position du cavalier
  @param[in] p - numero du pas
  @param[in,out] sols - la liste des solutions
  @pre pos est dans l'echiquier et non encore marque
*/
void dfsRecCV(Echiquier& cb, const CVDeplacements& delta,
  const Position& pos, unsigned p, list<Echiquier>& sols)
{
  // Marque la cellule
  cb.set(pos, p);

  // Est-ce une solution?
  if (cb.sizeN()*cb.sizeN() == p)
  {
    sols.push_back(cb);
  }

  // Sinon tente la traversee en chacune des positions possibles
  for (unsigned nv = 0; nv < delta.size(); ++nv)
  {
    Position nextpos = pos + delta[nv];
    if (cb.contient(nextpos) && cb.get(nextpos) == 0)
    {
      dfsRecCV(cb, delta, nextpos, p+1, sols);
    }
  }

  // Libere la cellule
  cb.set(pos, 0);
}

/**
  Parcours en profondeur, version recursive
  Lance la traversee a partir d'une position depart
  @param[in,out] cb - un Echiquier
  @param[in] delta - delta de deplacements
  @param[in] depart - Position de depart
  @param[out] sols - la liste des solutions
*/
void dfsCV(Echiquier& cb, const CVDeplacements& delta,
    const Position& depart, list<Echiquier>& sols)
{
  // Initialise l'echiquier
  cb.reset(0);

  // Lance la traversee
  sols.clear();
  if (cb.contient(depart))
  {
    dfsRecCV(cb, delta, depart, 1, sols);
  }
}

/**
  Affiche les solutions de la traversee
  @param[in] sols - la liste des Echiquier-solutions
*/
void afficherListeDFS(const list<Echiquier>& sols)
{
  cout<<"Traversee en profondeur: nombre de solutions = "<<sols.size()<<endl;
  int nsols;
  do{
    cout<<"Nombre de solutions a afficher? ";
    cin >> nsols;
  } while (nsols < 0 || sols.size()< nsols);
  list<Echiquier>::const_iterator it = sols.begin();
  for (int k = 1; k <= nsols; ++k, ++it)
  {
    afficherEchiquier(*it, 3);
  }
}

/**
  @test
*/
void test_DFS()
{
  // Saisit des donnees
  cout<<"Dimension n de l'echiquier? ";
  unsigned n;
  cin >> n;
  Echiquier cb(n);
  cout<<"Mouvements standard (0 == oui)? ";
  unsigned rep;
  cin >> rep;
  cout<<"Position (x,y) de depart? ";
  Position cell;
  cin >> cell;
  // Lance la traversee en profondeur
  list<Echiquier> sols;
  if (rep == 0)
  {
    dfsCV(cb, delta0, cell, sols);
  }
  else
  {
    dfsCV(cb, delta1, cell, sols);
  }
  // Affiche le resultat
  afficherListeDFS(sols);
}

/**
  Parcours en profondeur, version iterative
  @param[in,out] cb - un Echiquier
  @param[in] delta - delta de deplacements
  @param[in] depart - Position de depart
  @param[out] sols - la liste des Echiquier-solutions
*/
void dfsIterCV(Echiquier& cb, const CVDeplacements& delta,
    const Position& depart, list<Echiquier>& sols)
{
  // Reinitialise l'echiquier
  cb.reset(0);

  // Couplet sur la pile: (position, # direction)
  typedef pair<Position, unsigned> Couplet;
  stack<Couplet, vector<Couplet> > pile;

  // Demarre la traversee a partir de depart
  if (cb.contient(depart))
  {
    cb.set(depart, 1);
    pile.push( make_pair(depart, -1) );
  }

  // Traverse tantque possible
  while( not pile.empty() ) // && sols.size() < mxsols)
  {
    // Recupere les informations du sommet de la pile
    // pour continuer a traiter la position pos a partir
    // de la prochaine direction
    Position pos = pile.top().first;
    unsigned ndir = pile.top().second;

    // Cherche la prochaine position disponible issue
    // de pos selon la direction ndir suivante
    Position nextpos;
    while (++ndir < delta.size())
    {
      nextpos = pos + delta[ndir];
      if (cb.contient(nextpos) && cb.get(nextpos) == 0)
      {
        break;
      }
    }

    // En cas de recherche frutueuse, continue la traversee a
    // partir de la position nextpos. Sinon cas d'echec: libere
    // la position pos et revient en arriere (backtracking)
    if (ndir < delta.size())
    {
      // Notifie la direction
      pile.top().second = ndir;

      // Valeur de la cellule
      int n = cb.get(pos) + 1;

      // Marque la cellule
      cb.set(nextpos, n);

      // Est-ce une solution? Si oui
      if (n >= cb.sizeN()*cb.sizeN())
      {
        // Enfile la solution
        sols.push_back(cb);

//        // Continue la recherche?
//        if (sols.size() >= mxsols)
//        {
//          break;
//        }
      }

      // Empile le couplet suivant
      pile.push(make_pair(nextpos, -1));
    }
    else
    {
      // Ici: cul de sac: depile le couplet
      pile.pop();

      // Libere la position
      cb.set(pos, 0);
    }
  }
  // Fin de traversee: si la pile est vide, la traversee est
  // un parcours exhaustif: le systeme detruit la pile:
}

/**
  @test
*/
void test_DFSIter()
{
  // Saisit des donnees
  cout<<"Dimension n de l'echiquier? ";
  unsigned n;
  cin >> n;
  Echiquier cb(n);
  cout<<"Mouvements standard (0 == oui)? ";
  unsigned rep;
  cin >> rep;
  cout<<"Position (x,y) de depart? ";
  Position cell;
  cin >> cell;
  // Lance la traversee en profondeur
  list<Echiquier> sols;
  if (rep == 0)
  {
    dfsIterCV(cb, delta0, cell, sols);
  }
  else
  {
    dfsIterCV(cb, delta1, cell, sols);
  }
  // Affiche le resultat
  afficherListeDFS(sols);
}

/**
  Reinitialise l'echiquier et le tableau des accessibilites
  @param[out] cb - un Echiquier
  @param[out] hb - Echiquier des accessibilites
  @param[in] delta - delta de deplacements
*/
void initHeuristiq(Echiquier& cb, Echiquier& hb, const CVDeplacements& delta)
{
  // Reinitialise l'echiquier de parcours du cavalier
  cb.reset(0);

  // Reinitialise l'accessibilite de chacune des positions
  hb.reset(0);
  for (unsigned y = 0; y < cb.sizeN(); ++y)
  {
    for (unsigned x = 0; x < cb.sizeN(); ++x)
    {
      Position pos(x, y);
      unsigned a = 0;
      for (unsigned nv = 0; nv < delta.size(); ++nv)
      {
        Position nextpos = pos + delta[nv];
        if ( cb.contient(nextpos) )
        {
          ++a;
        }
      }
      hb.set(pos, a);
    }
  }
}

/**
  Lance la traversee heuristique
  @param[in,out] cb - un Echiquier
  @param[in] delta - delta de deplacements
  @param[in] depart - Position de depart
  @param[in,out] hb - Echiquier des accessibilites
*/
void heuristiqCV(Echiquier& cb, const CVDeplacements& delta,
    const Position& depart, Echiquier& hb)
{
  // Reinitialise les donnees
  initHeuristiq(cb, hb, delta);

  // Demarre la traversee a partir de depart
  if ( !cb.contient(depart) )
  {
    return;
  }

  // Traverse tantque possible
  const unsigned NULL_ACCESS = delta.size()+1;
  // position courante
  Position pos = depart;
  // pas courant
  int step = 0;
  while(true)
  {
    // Marque la position
    cb.set(pos, ++step);
    hb.set(pos, NULL_ACCESS);

    // Recherche la prochaine position
    Position nextpos;
    // Et sa valeur d'accessibilite
    unsigned nacces = NULL_ACCESS;
    for (unsigned nv = 0; nv < delta.size(); ++nv)
    {
      // Calcule la position destination
      Position nvpos = pos + delta[nv];

      // Si dans l'echiquier et non marquee
      if ( cb.contient(nvpos) && cb.get(nvpos) == 0 )
      {
        // Decremente l'accessibilite de nvpos puis
        // teste si c'est la suivante
        unsigned val = hb.get(nvpos);
        hb.set(nvpos, --val);
        if (val < nacces)
        {
          nacces = val;
          nextpos = nvpos;
        }
      }
    }

    // En cas d'echec: fin de la traversee
    if (NULL_ACCESS == nacces)
    {
      break;
    }

    // Sinon, on se deplace sur la prochaine position
    pos = nextpos;
  }
}

/**
  Affiche le resultat
  @param[in] cb - un Echiquier
  @param[in] hb - Echiquier des accessibilites
*/
void afficherHeuristiq(const Echiquier& cb, const Echiquier& hb)
{
  // Recherche du dernier pas
  int step = 0;
  for (unsigned y = 0; y < cb.sizeN(); ++y)
  {
    for (unsigned x = 0; x < cb.sizeN(); ++x)
    {
      if (cb.get(x, y) > step)
      {
        step = cb.get(x, y);
      }
    }
  }
  if (step == cb.sizeN() * cb.sizeN())
  {
    cout<<"SOLUTION trouvee"<<endl;
  }
  else
  {
    cout<<"Dernier pas: "<<step<<endl;
  }
  cout<<"Echiquier final: "<<endl;
  afficherEchiquier(cb, 3);
  cout<<" Dernier HB: "<<endl;
  afficherEchiquier(hb, 3);
}

/**
  @test
*/
void test_Heuristiq()
{
  // Saisit des donnees
  cout<<"Dimension n de l'echiquier? ";
  unsigned n;
  cin >> n;
  Echiquier cb(n);
  cout<<"Mouvements standard (0 == oui)? ";
  unsigned rep;
  cin >> rep;
  cout<<"Position (x,y) de depart? ";
  Position cell;
  cin >> cell;
  // Lance la traversee heuristique
  Echiquier hb(n);
  if (rep == 0)
  {
    heuristiqCV(cb, delta0, cell, hb);
  }
  else
  {
    heuristiqCV(cb, delta1, cell, hb);
  }
  // Affiche le resultat
  afficherHeuristiq(cb, hb);
}

int main()
{
  test_BFS();
  test_DFS();
  test_DFSIter();
  test_Heuristiq();
}

